#include "cutil.h"
#include "cutil_inline.h"
#include <stdio.h>
#include <string.h>
#include "spmvELLRTdp.h"

void cudaInit(int dev)
{
    int deviceCount;                                                         
    cudaGetDeviceCount(&deviceCount);                
    if (deviceCount == 0) {                                                  
        fprintf(stderr, "cutil error: no devices supporting CUDA.\n");       
        exit(EXIT_FAILURE);                                                  
    }                                                                        
    
    if (dev < 0) dev = cutGetMaxGflopsDeviceId();

    if (dev > deviceCount-1) dev = deviceCount - 1;                          
    cudaDeviceProp deviceProp;                                               
    cudaGetDeviceProperties(&deviceProp, dev);       
    if (deviceProp.major < 1) {                                              
        fprintf(stderr, "cutil error: device does not support CUDA.\n");     
        exit(EXIT_FAILURE);                                                  
    }                                                                        
    fprintf(stderr, "Using device %d: %s\n", dev, deviceProp.name);       
    cudaSetDevice(dev);                                 
}


int main(int argc,char *argv[]){

double *A; 	 	 /** Sparse matrix **/
double *v;               /** Vector to multiply **/
double *u;               /** Result **/
unsigned int *index;    /** Indices matrix **/
unsigned int *col;      /** Row-length **/ 

double       *BA;
unsigned int *BIndex;
unsigned int *BCol;

int blocksize,device,symmetric,threads,ones,texture;

char *file;       /* Sparse matrix file */
FILE *pf;         /* Pointer to the file */
char line[200];   /* Complete line of the file */
float value;      /* Value of the element i,j  */

  
           file=argv[1];
      blocksize=atoi(argv[2]);   
        threads=atoi(argv[3]);  
         device=atoi(argv[4]);
        texture=atoi(argv[5]);

  cudaInit(device);

  int nrows,ncols,nzeros,alignment,nmaxblocks;
  size_t size_a,size_v,size_u,size_i,size_c;
  int i,j,max,min;

  /** Get the sparse matrix from the input file **/

    if((pf = fopen(file,"r"))==NULL){
  	printf("Can't open the file: %s",file);
  	exit(1);
  }

        char banner[64];
        char mtx[64];   
        char crd[64];
        char data_type[64];
        char storage_scheme[64];
 
  	fgets(line,200,pf);                 /* First line: matrix name */
    
        sscanf(line,"%s %s %s %s %s", banner, mtx, crd, data_type, storage_scheme);    

        ones=0;symmetric=0;

        if (strcmp(data_type,"pattern")==0) ones=1;    
     
        if (strcmp(storage_scheme,"symmetric")==0) symmetric=1;

        fscanf(pf,"%d %d %d",&nrows,&ncols,&nzeros);  /* Second line: rows, cols, non zeros=entries */
     
        printf("\nSym: %d. Pattern: %d. Nrows: %d. Ncols: %d. Entries: %d",symmetric,ones,nrows,ncols,nzeros);
        printf("\n");
        printf("\n");


        /* First step. Calculate the max column */

        size_c=nrows*sizeof(unsigned int);
        col=(unsigned int *)malloc(size_c);

        for(i=0;i<nrows;i++)
          col[i]=0;

        while (!feof(pf)){
	  if (ones==0)fscanf(pf,"%d %d %f",&i,&j,&value);
          else if(ones==1)fscanf(pf,"%d %d",&i,&j);
          else fscanf(pf,"%f %d %d",&value,&i,&j);
       
          if(ones<2){
  	   i--;  /*Begin from 1*/ 
	   j--;
          }

	  col[i]++;
          if((symmetric==1)&&(i!=j))col[j]++;
        }
       
	fclose(pf);

        max=0;
        min=ncols;
        for(i=0;i<nrows;i++){
          if(col[i]>max)max=col[i];
          if(col[i]<min)min=col[i];
        }



        size_a=max*nrows*sizeof(double);
        size_i=max*nrows*sizeof(unsigned int);
               
        A=(double *)malloc(size_a);
        index=(unsigned int *)malloc(size_i);
        
        /* Initalization of A,index with zeros */

        for(i=0;i<nrows*max;i++){
               A[i]=0.0; 
               index[i]=0; 
               if (i<nrows)col[i]=0;
        }



        /* Second step: the values of the matrix */

        if((pf = fopen(file,"r"))==NULL){
   	  printf("Can't open the file: %s",file);
   	  exit(1);
        }
  
  	fgets(line,200,pf);              
        fscanf(pf,"%d %d %d",&nrows,&ncols,&nzeros);                

        nzeros=0;

        value=1.0;
        while (!feof(pf)){
               if(ones==0)fscanf(pf,"%d %d %f",&i,&j,&value);
          else if(ones==1)fscanf(pf,"%d %d",&i,&j);
          else fscanf(pf,"%f %d %d",&value,&i,&j);
       
          if(ones<2){
  	   i--;  /*Begin from 1 */
	   j--;
          }

	  A[i*max+col[i]]=value;
          index[i*max+col[i]]=j;
          col[i]++;
          nzeros++;
          if((symmetric==1)&&(i!=j)){
             
                A[j*max+col[j]]=value;
                index[j*max+col[j]]=i;
                col[j]++;
                nzeros++;
          }


	}
	fclose(pf);




        size_v=ncols*sizeof(double);
        size_u=nrows*sizeof(double);
        
        v=(double *)malloc(size_v);
        u=(double *)malloc(size_u);



        srand48(time(NULL));        
        for(i=0;i<nrows;i++){
         if (i<ncols)v[i]=drand48();
         u[i]=0.0;
        }


        /* Matrix BA, BIndex, BCol */

        alignment=ceil((float)(nrows*threads)/128)*128;
        nmaxblocks=ceil((float)max/threads);

        size_t size_ba=nmaxblocks*alignment*sizeof(double);
        size_t size_bi=nmaxblocks*alignment*sizeof(unsigned int);

        BA=(double *)malloc(size_ba);       /* Matrix in ELLRT format */
        BIndex=(unsigned int *)malloc(size_bi); /* Index columns */
        BCol=(unsigned int *)malloc(size_c); /* Row length */

        for(i=0;i<nmaxblocks*alignment;i++){
          BA[i]=0.0;
          BIndex[i]=0;
          if (i<nrows)BCol[i]=ceil((float)col[i]/threads);
        }

        int k;
 
        for(i=0;i<nrows;i++){
          j=0;
          k=0;
          while(j<col[i]){
           for(int m=0;m<threads;m++){
             if(j<col[i]){
               BA[k*alignment+(i*threads)+m]=A[i*max+j];
               BIndex[k*alignment+(i*threads)+m]=index[i*max+j];
             }
             j++;
           }
           k++;
          }
        }


   /* Device's structures */

   double *d_BA;cudaMalloc((void**)&d_BA,size_ba);
   double *d_v;cudaMalloc((void**)&d_v,size_v);
   double *d_u;cudaMalloc((void**)&d_u,size_u);
   unsigned int *d_BIndex;cudaMalloc((void**)&d_BIndex,size_bi);
   unsigned int *d_BCol;cudaMalloc((void**)&d_BCol,size_c);


   /* Copy data to GPU */
   cudaMemcpy(d_BA,BA,size_ba,cudaMemcpyHostToDevice);   
   cudaMemcpy(d_BIndex,BIndex,size_bi,cudaMemcpyHostToDevice);
   cudaMemcpy(d_BCol,BCol,size_c,cudaMemcpyHostToDevice);

   cudaMemcpy(d_v,v,size_v,cudaMemcpyHostToDevice);   
   cudaMemcpy(d_u,u,size_u,cudaMemcpyHostToDevice);   


   /** Library **/
   
   spmv_ELLRTdp(d_BA,d_v,d_u,d_BIndex,d_BCol,nrows,alignment,blocksize,threads,texture);


  /* Copy data from GPU to host */ 

   cudaMemcpy(u,d_u,size_u,cudaMemcpyDeviceToHost);       


   // Free memory in the GPU

   cudaFree(d_BA);
   cudaFree(d_v);
   cudaFree(d_u);
   cudaFree(d_BIndex);
   cudaFree(d_BCol);

   // Free memory in the host

   free(A);
   free(index);
   free(BA);
   free(BIndex);
   free(col);
   free(BCol);
   free(v);
   free(u);



     printf("\n***");
     printf("\n   Algorithm: ELLRT double precision. T = %d. BS = %d. Alignment = %d",threads,blocksize,alignment);
     printf("\n   Memory: %2.4f MB. Matrix name: %s. NonZeros: %d %dx%d",(size_ba+size_v+size_u+size_bi+size_c)/1048576.0,file,nzeros,nrows,ncols);
     printf("\n***");
     printf("\n");

}






